from copy import deepcopy
from typing import Dict
import numpy as np

from skellytracker.trackers.base_tracker.base_recorder import BaseRecorder
from skellytracker.trackers.base_tracker.tracked_object import TrackedObject
from skellytracker.trackers.mediapipe_tracker.mediapipe_model_info import (
    MediapipeModelInfo,
)


class MediapipeHolisticRecorder(BaseRecorder):
    def record(self, tracked_objects: Dict[str, TrackedObject]) -> None:
        self.recorded_objects.append(
            [
                deepcopy(tracked_objects[tracked_object_name])
                for tracked_object_name in MediapipeModelInfo.tracked_object_names
            ]
        )

    def process_tracked_objects(self, **kwargs) -> np.ndarray:
        image_size = kwargs.get("image_size")
        if image_size is None:
            raise ValueError(
                f"image_size must be provided to process tracked objects from {__class__.__name__}"
            )
        self.recorded_objects_array = np.zeros(
            (
                len(self.recorded_objects),
                MediapipeModelInfo.num_tracked_points,
                3,
            )
        )

        for i, recorded_object_list in enumerate(self.recorded_objects):
            landmark_number = 0
            for recorded_object in recorded_object_list:
                if recorded_object.extra["landmarks"] is not None:
                    for landmark_data in recorded_object.extra["landmarks"].landmark:
                        self.recorded_objects_array[i, landmark_number, 0] = (
                            landmark_data.x * image_size[0]
                        )
                        self.recorded_objects_array[i, landmark_number, 1] = (
                            landmark_data.y * image_size[1]
                        )
                        self.recorded_objects_array[i, landmark_number, 2] = (
                            landmark_data.z * image_size[0]
                        )  # * image width per mediapipe docs
                        landmark_number += 1
                else:
                    if recorded_object.object_id == "pose_landmarks":
                        number = MediapipeModelInfo.num_tracked_points_body
                    elif recorded_object.object_id == "face_landmarks":
                        number = MediapipeModelInfo.num_tracked_points_face
                    elif recorded_object.object_id == "left_hand_landmarks":
                        number = MediapipeModelInfo.num_tracked_points_left_hand
                    else:
                        number = MediapipeModelInfo.num_tracked_points_right_hand
                    for _ in range(number):
                        self.recorded_objects_array[i, landmark_number, :] = np.nan
                        landmark_number += 1

        return self.recorded_objects_array



import math
import pytest
import numpy as np


from skellytracker.trackers.yolo_tracker.yolo_model_info import YOLOModelInfo
from skellytracker.trackers.yolo_tracker.yolo_tracker import YOLOPoseTracker


@pytest.mark.usefixtures("test_image")
def test_process_image(test_image):
    tracker = YOLOPoseTracker(model_size="nano")
    tracked_objects = tracker.process_image(test_image)

    assert len(tracked_objects) == 1
    tracked_person = tracked_objects["tracked_person"]
    assert tracked_person.pixel_x is not None
    assert tracked_person.pixel_y is not None
    assert math.isclose(tracked_person.pixel_x, 266.48523, rel_tol=1e-2)
    assert math.isclose(tracked_person.pixel_y, 273.92798, rel_tol=1e-2)

    landmarks = tracked_person.extra["landmarks"]
    assert landmarks is not None
    assert landmarks.shape == (1, YOLOModelInfo.num_tracked_points, 2)

    expected_results = np.array(
        [
            [
                [392.56927490234375, 140.40118408203125],
                [414.94940185546875, 132.90655517578125],
                [386.4353332519531, 125.51483154296875],
                [446.2061767578125, 157.98883056640625],
                [373.6619873046875, 138.93646240234375],
                [453.78662109375, 265.081787109375],
                [317.9375305175781, 231.9653778076172],
                [465.893310546875, 396.12274169921875],
                [220.12176513671875, 325.96636962890625],
                [465.23358154296875, 499.0487365722656],
                [142.17066955566406, 407.7397155761719],
                [352.325439453125, 468.44671630859375],
                [268.8867492675781, 448.7227783203125],
                [310.899658203125, 630.5478515625],
                [227.7810821533203, 617.9011840820312],
                [269.08587646484375, 733.4285888671875],
                [213.1557159423828, 741.54541015625],
            ]
        ]
    )

    assert np.allclose(landmarks, expected_results)


@pytest.mark.usefixtures("test_image")
def test_annotate_image(test_image):
    tracker = YOLOPoseTracker(model_size="nano")
    tracker.process_image(test_image)

    assert tracker.annotated_image is not None


@pytest.mark.usefixtures("test_image")
def test_record(test_image):
    tracker = YOLOPoseTracker(model_size="nano")
    tracked_objects = tracker.process_image(test_image)
    tracker.recorder.record(tracked_objects=tracked_objects)
    assert len(tracker.recorder.recorded_objects) == 1

    processed_results = tracker.recorder.process_tracked_objects()
    assert processed_results is not None
    assert processed_results.shape == (1, YOLOModelInfo.num_tracked_points, 3)

    expected_results = np.array(
        [
            [
                [392.56927490234375, 140.40118408203125, np.nan],
                [414.94940185546875, 132.90655517578125, np.nan],
                [386.4353332519531, 125.51483154296875, np.nan],
                [446.2061767578125, 157.98883056640625, np.nan],
                [373.6619873046875, 138.93646240234375, np.nan],
                [453.78662109375, 265.081787109375, np.nan],
                [317.9375305175781, 231.9653778076172, np.nan],
                [465.893310546875, 396.12274169921875, np.nan],
                [220.12176513671875, 325.96636962890625, np.nan],
                [465.23358154296875, 499.0487365722656, np.nan],
                [142.17066955566406, 407.7397155761719, np.nan],
                [352.325439453125, 468.44671630859375, np.nan],
                [268.8867492675781, 448.7227783203125, np.nan],
                [310.899658203125, 630.5478515625, np.nan],
                [227.7810821533203, 617.9011840820312, np.nan],
                [269.08587646484375, 733.4285888671875, np.nan],
                [213.1557159423828, 741.54541015625, np.nan],
            ]
        ]
    )
    assert np.allclose(
        processed_results[:, :, :2], expected_results[:, :, :2], atol=1e-2
    )
    assert np.isnan(processed_results[:, :, 2]).all()


class MockKeypoints:
    xy: list = []


class MockResults:
    keypoints: MockKeypoints = MockKeypoints()


def test_unpack_empty_results():
    tracker = YOLOPoseTracker(model_size="nano")
    results = [MockResults()]
    tracker.unpack_results(results=results)

    tracked_person = tracker.tracked_objects["tracked_person"]
    assert tracked_person.pixel_x is None
    assert tracked_person.pixel_y is None
    assert tracked_person.extra["landmarks"].shape == (
        1,
        YOLOModelInfo.num_tracked_points,
        2,
    )
    assert np.isnan(tracked_person.extra["landmarks"][0, :, 0]).all()
    assert np.isnan(tracked_person.extra["landmarks"][0, :, 1]).all()



import pytest
import numpy as np


from skellytracker.trackers.yolo_object_tracker.yolo_object_tracker import (
    YOLOObjectTracker,
)


@pytest.mark.usefixtures("test_image")
def test_process_image_person_only(test_image):
    tracker = YOLOObjectTracker(model_size="nano", person_only=True)
    tracked_objects = tracker.process_image(test_image)

    assert len(tracked_objects) == 1
    assert tracked_objects["object"] is not None
    assert tracked_objects["object"].extra["boxes_xyxy"] is not None
    assert np.allclose(
        tracked_objects["object"].extra["boxes_xyxy"],
        [90.676, 96.981, 493.54, 812.03],
        atol=1e-2,
    )
    assert tracked_objects["object"].extra["original_image_shape"] == (1280, 720)


@pytest.mark.usefixtures("test_image")
def test_annotate_image(test_image):
    tracker = YOLOObjectTracker()
    tracker.process_image(test_image)

    assert tracker.annotated_image is not None


@pytest.mark.usefixtures("test_image")
def test_record(test_image):
    tracker = YOLOObjectTracker(model_size="nano", person_only=True)
    tracked_objects = tracker.process_image(test_image)
    tracker.recorder.record(tracked_objects=tracked_objects)
    assert len(tracker.recorder.recorded_objects) == 1

    processed_results = tracker.recorder.process_tracked_objects()
    assert processed_results is not None
    assert processed_results.shape == (1, 4)

    assert np.allclose(processed_results, [90.676, 96.981, 493.54, 812.03], atol=1e-2)



import logging
import numpy as np
from multiprocessing import Pool, cpu_count
from pathlib import Path
from typing import Optional
from pydantic import BaseModel


from skellytracker.system.constants import BASE_2D_FILE_NAME
from skellytracker.trackers.base_tracker.base_tracker import BaseTracker
from skellytracker.trackers.base_tracker.model_info import ModelInfo
from skellytracker.trackers.bright_point_tracker.brightest_point_tracker import (
    BrightestPointTracker,
)
from skellytracker.utilities.get_video_paths import get_video_paths

try:
    from skellytracker.trackers.yolo_mediapipe_combo_tracker.yolo_mediapipe_combo_tracker import (
        YOLOMediapipeComboTracker,
    )
except ModuleNotFoundError:
    print(
        "\n\nTo use yolo_mediapipe_combo_tracker, install skellytracker[yolo, mediapipe]\n\n"
    )
try:
    from skellytracker.trackers.yolo_tracker.yolo_tracker import YOLOPoseTracker
    from skellytracker.trackers.yolo_tracker.yolo_model_info import YOLOTrackingParams
except ModuleNotFoundError:
    print("To use yolo_tracker, install skellytracker[yolo]")
try:
    from skellytracker.trackers.mediapipe_tracker.mediapipe_holistic_tracker import (
        MediapipeHolisticTracker,
    )
    from skellytracker.trackers.mediapipe_tracker.mediapipe_model_info import (
        MediapipeTrackingParams,
    )
except ModuleNotFoundError:
    print("To use mediapipe_holistic_tracker, install skellytracker[mediapipe]")

logger = logging.getLogger(__name__)

try:
    from skellytracker.trackers.openpose_tracker.openpose_tracker import (
        OpenPoseTracker,
    )
except ModuleNotFoundError:
    print("To use openpose_tracker, install skellytracker[openpose]")

logger = logging.getLogger(__name__)


def process_folder_of_videos(
    model_info: ModelInfo,
    tracking_params: BaseModel,
    synchronized_video_path: Path,
    output_folder_path: Optional[Path] = None,
    annotated_video_path: Optional[Path] = None,
    num_processes: Optional[int] = None,
) -> np.ndarray:
    """
    Process a folder of synchronized videos with the given tracker.
    Tracked data will be saved to a .npy file with the shape (numCams, numFrames, numTrackedPoints, pixelXYZ).

    :param model_info: Model info for tracker.
    :param tracking_params: Tracking parameters to use.
    :param synchronized_video_path: Path to folder of synchronized videos.
    :param output_folder_path: Path to save tracked data to.
    :param annotated_video_path: Path to save annotated videos to.
    :param num_processes: Number of processes to use, 1 to disable multiprocessing.
    :return: Array of tracking data
    """
    video_paths = get_video_paths(synchronized_video_path)

    if num_processes is None:
        num_processes = min((cpu_count() - 1), len(video_paths))
    else:
        num_processes = min(num_processes, len(video_paths), cpu_count() - 1)

    file_name = model_info.name + "_" + BASE_2D_FILE_NAME
    synchronized_video_path = Path(synchronized_video_path)
    if output_folder_path is None:
        output_folder_path = (
            synchronized_video_path.parent / "output_data" / "raw_data" / file_name
        )
    else:
        output_folder_path = Path(output_folder_path) / file_name
    if not output_folder_path.exists():
        output_folder_path.parent.mkdir(parents=True, exist_ok=True)

    if annotated_video_path is None:
        annotated_video_path = synchronized_video_path.parent / "annotated_videos"
    if not annotated_video_path.exists():
        annotated_video_path.mkdir(parents=True, exist_ok=True)

    tasks = [
        (model_info.tracker_name, tracking_params, video_path, annotated_video_path)
        for video_path in video_paths
    ]

    if num_processes > 1:
        logging.info("Using multiprocessing to run pose estimation")
        with Pool(processes=num_processes) as pool:
            array_list = pool.starmap(process_single_video, tasks)
    else:
        array_list = []
        for task in tasks:
            array_list.append(process_single_video(*task))

    combined_array = np.stack(array_list)

    logger.info(f"Shape of output array: {combined_array.shape}")
    np.save(output_folder_path, combined_array)

    return combined_array


def process_single_video(
    tracker_name: str,
    tracking_params: BaseModel,
    video_path: Path,
    annotated_video_path: Path,
) -> Optional[np.ndarray]:
    """
    Process a single video with the given tracker.
    Tracked data will be saved to a .npy file with the shape (numCams, numFrames, numTrackedPoints, pixelXYZ).

    :param tracker_name: Tracker to use.
    :param tracking_params: Tracking parameters to use.
    :param video_path: Path to video.
    :param annotated_video_path: Path to save annotated video to.
    :return: Array of tracking data
    """

    if tracker_name == "OpenPoseTracker":
        video_name = video_path.stem + "_openpose.avi"
    else:
        video_name = (
            video_path.stem + "_mediapipe.mp4"
        )  # TODO: fix it so blender output doesn't require mediapipe addendum here

    tracker = get_tracker(tracker_name=tracker_name, tracking_params=tracking_params)
    logger.info(
        f"Processing video: {video_name} with tracker: {tracker.__class__.__name__}"
    )
    output_array = tracker.process_video(
        input_video_filepath=video_path,
        output_video_filepath=annotated_video_path / video_name,
        save_data_bool=False,
    )  # TODO: raise a custom error here if output_array is None?
    return output_array


def get_tracker(tracker_name: str, tracking_params: BaseModel) -> BaseTracker:
    """
    Returns a tracker object based on the given tracker_type and tracking_params.

    :param tracker_type (str): The type of tracker to be created.
    :param tracking_params (BaseModel): The tracking parameters to be used for creating the tracker.
    :return BaseTracker: The tracker object based on the given tracker_type and tracking_params.
    :raise ValueError: If an invalid tracker_type is provided.
    """
    if tracker_name == "MediapipeHolisticTracker":
        tracker = MediapipeHolisticTracker(
            model_complexity=tracking_params.mediapipe_model_complexity,
            min_detection_confidence=tracking_params.min_detection_confidence,
            min_tracking_confidence=tracking_params.min_tracking_confidence,
            static_image_mode=tracking_params.static_image_mode,
        )

    elif tracker_name == "YOLOMediapipeComboTracker":
        tracker = YOLOMediapipeComboTracker(
            model_size=tracking_params.yolo_model_size,
            model_complexity=tracking_params.mediapipe_model_complexity,
            min_detection_confidence=tracking_params.min_detection_confidence,
            min_tracking_confidence=tracking_params.min_tracking_confidence,
            static_image_mode=True,  # yolo cropping must be run with static image mode due to changing size of bounding boxes
            bounding_box_buffer_percentage=tracking_params.bounding_box_buffer_percentage,
            buffer_size_method=tracking_params.buffer_size_method,
        )

    elif tracker_name == "YOLOPoseTracker":
        tracker = YOLOPoseTracker(
            model_size="medium",
        )

    elif tracker_name == "BrightestPointTracker":
        tracker = BrightestPointTracker()

    elif tracker_name == "OpenPoseTracker":
        tracker = OpenPoseTracker(
            openpose_root_folder_path=tracking_params.openpose_root_folder_path,
            output_json_folder_path=tracking_params.output_json_path,
            net_resolution=tracking_params.net_resolution,
            number_people_max=tracking_params.number_people_max,
            track_faces=tracking_params.track_face,
            track_hands=tracking_params.track_hands,
            output_resolution=tracking_params.output_resolution,
        )

    else:
        raise ValueError("Invalid tracker type")

    return tracker


def get_tracker_params(tracker_name: str) -> BaseModel:
    if tracker_name == "MediapipeHolisticTracker":
        return MediapipeTrackingParams()
    elif tracker_name == "YOLOMediapipeComboTracker":
        return YOLOTrackingParams()  # TODO: figure out how to reference both tracking params in a stable way
    elif tracker_name == "YOLOPoseTracker":
        return YOLOTrackingParams()
    elif tracker_name == "BrightestPointTracker":
        return BaseModel()
    elif tracker_name == "OpenPoseTracker":
        raise ValueError(
            "OpenPoseTracker requires explicitly setting the OpenPose root folder path and output json path, please provide tracking params directly"
        )
    else:
        raise ValueError("Invalid tracker type")


if __name__ == "__main__":
    from skellytracker.trackers.mediapipe_tracker.mediapipe_model_info import MediapipeModelInfo

    synchronized_video_path = Path(
        "/Your/Path/To/freemocap_data/recording_sessions/freemocap_sample_data/synchronized_videos"
    )

    tracker_name = "YOLOMediapipeComboTracker"
    num_processes = None

    process_folder_of_videos(
        model_info=MediapipeModelInfo(),
        tracking_params=get_tracker_params(tracker_name=tracker_name),
        synchronized_video_path=synchronized_video_path,
        num_processes=num_processes,
    )



import subprocess
from pathlib import Path
from typing import Optional, Union
from skellytracker.trackers.base_tracker.base_tracker import BaseCumulativeTracker
from skellytracker.trackers.openpose_tracker.openpose_recorder import OpenPoseRecorder


class OpenPoseTracker(BaseCumulativeTracker):
    def __init__(
        self,
        openpose_root_folder_path: Union[str, Path],
        output_json_folder_path: Optional[Union[str, Path]] = None,
        net_resolution: str = "-1x320",
        number_people_max: int = 1,
        track_hands: bool = True,
        track_faces: bool = True,
        output_resolution: str = "-1x-1",
    ):
        """
        Initialize the OpenPoseTracker.

        :param recorder: An instance of OpenPoseRecorder for handling the output.
        :param openpose_root_folder_path: Path to the OpenPose root folder.
        :param output_json_folder_path: Path to the output JSON folder.
        :param net_resolution: Network resolution for OpenPose processing.
        :param number_people_max: Maximum number of people to detect.
        :param track_hands: Whether to track hands.
        :param track_faces: Whether to track faces.
        :param output_resolution: Output resolution for video.
        """
        super().__init__(
            tracked_object_names=[],
            recorder=OpenPoseRecorder(
                track_hands=track_hands,
                track_faces=track_faces,
            ),
            track_hands=track_hands,
            track_faces=track_faces,
        )
        self.openpose_root_folder_path = Path(openpose_root_folder_path)
        self.output_json_folder_path = output_json_folder_path
        self.net_resolution = net_resolution
        self.number_people_max = number_people_max
        self.track_hands = track_hands
        self.track_faces = track_faces
        self.output_resolution = output_resolution

    def set_track_hands(self, track_hands: bool):
        self._track_hands = track_hands
        self.recorder.track_hands = track_hands

    def set_track_faces(self, track_faces: bool):
        self._track_faces = track_faces
        self.recorder.track_faces = track_faces

    def set_json_output_path(self, output_json_folder_path: Union[str, Path]):
        self.output_json_folder_path = Path(output_json_folder_path)

    def process_video(
        self,
        input_video_filepath: Union[str, Path],
        output_video_filepath: Union[str, Path],
        save_data_bool: bool = False,
        use_tqdm: bool = True,  # TODO: this is unused, replace with an openpose flag or remove
        **kwargs,
    ):
        """
        Run the OpenPose demo on a video file to generate JSON outputs
        in a unique directory for each video.

        :param input_video_filepath: Path to the input video file.
        :param output_video_filepath: Path to the output video file.
        :param save_data_bool: Whether to save the data.
        :param use_tqdm: Whether to use tqdm progress bar.
        :return: The output array, or None if recorder isn't initialized in tracker.
        """
        # Extract video name without extension to use as a unique folder name
        video_name = Path(input_video_filepath).stem

        if self.output_json_folder_path is None:
            self.output_json_folder_path = Path(input_video_filepath).parent.parent / "output_data" / "raw_data" / "openpose_jsons"

        Path(self.output_json_folder_path).mkdir(parents=True, exist_ok=True)

        unique_json_output_path = Path(self.output_json_folder_path) / video_name
        unique_json_output_path.mkdir(parents=True, exist_ok=True)

        # Full path to the OpenPose executable
        openpose_executable_path = (
            self.openpose_root_folder_path / "bin" / "OpenPoseDemo.exe"
        )

        openpose_command = [
            str(openpose_executable_path),  # Full path to the OpenPose executable
            "--video",
            str(input_video_filepath),
            "--write_json",
            str(unique_json_output_path),
            "--net_resolution",
            str(self.net_resolution),
            "--number_people_max",
            str(self.number_people_max),
            "--write_video",
            str(output_video_filepath),
            "--output_resolution",
            str(self.output_resolution),
        ]

        if self.track_hands:
            openpose_command.append("--hand")
        if self.track_faces:
            openpose_command.append("--face")

        # Update the subprocess command to use the unique output directory
        try:
            subprocess.run(  # noqa: S603
                openpose_command,
                shell=False,
                cwd=self.openpose_root_folder_path,  # Set the current working directory for the subprocess
                check=True,
            )
        except subprocess.CalledProcessError as e:
            print(f"Error: {e}")
            return None

        if self.recorder is not None:
            output_array = self.recorder.process_tracked_objects(
                output_json_path=unique_json_output_path
            )
            if save_data_bool:
                self.recorder.save(
                    file_path=str(Path(input_video_filepath).with_suffix(".npy"))
                )
        else:
            output_array = None

        return output_array


if __name__ == "__main__":
    # Example usage
    openpose_root_folder_path = r"C:\openpose"
    input_video_filepath = r'C:\path\to\input\video.mp4'
    output_video_filepath = r'C:\path\to\output\video.avi'

    tracker = OpenPoseTracker(
        openpose_root_folder_path=str(openpose_root_folder_path),
    )
    tracker.process_video(input_video_filepath, output_video_filepath)



from typing import Dict
import numpy as np

from skellytracker.trackers.base_tracker.base_recorder import BaseRecorder
from skellytracker.trackers.base_tracker.tracked_object import TrackedObject


class BrightestPointRecorder(BaseRecorder):
    def record(self, tracked_objects: Dict[str, TrackedObject]) -> None:
        self.recorded_objects.append(
            [
                (tracked_object.pixel_x, tracked_object.pixel_y)
                for tracked_object in tracked_objects.values()
                if "brightest_point" in tracked_object.object_id
            ]
        )

    def process_tracked_objects(self, **kwargs) -> np.ndarray:
        num_frames = len(self.recorded_objects)
        num_points = len(self.recorded_objects[0]) if num_frames > 0 else 0

        self.recorded_objects_array = np.zeros((num_frames, num_points, 2))
        for i, recorded_object in enumerate(self.recorded_objects):
            for j, (pixel_x, pixel_y) in enumerate(recorded_object):
                self.recorded_objects_array[i, j, 0] = pixel_x
                self.recorded_objects_array[i, j, 1] = pixel_y

        return self.recorded_objects_array



from typing import Dict, List, Optional, Union


class ModelInfo(dict):
    name: str
    tracker_name: str
    landmark_names: List[str]
    num_tracked_points: int
    tracked_object_names: Optional[list] = None
    virtual_markers_definitions: Optional[Dict[str, Dict[str, List[Union[str, float]]]]] = None
    segment_connections: Optional[Dict[str, Dict[str, str]]] = None
    center_of_mass_definitions: Optional[Dict[str, Dict[str, float]]] = None
    joint_hierarchy: Optional[Dict[str, List[str]]] = None



import logging
import logging.handlers
import sys
from logging.config import dictConfig
from typing import Optional


DEFAULT_LOGGING = {"version": 1, "disable_existing_loggers": False}


def get_logging_handlers(log_file_path: Optional[str] = ""):
    dictConfig(DEFAULT_LOGGING)

    default_formatter = logging.Formatter(
        "[%(asctime)s.%(msecs)04d] [%(levelname)8s] [%(name)s] [%(funcName)s():%(lineno)s] [PID:%(process)d "
        "TID:%(thread)d] %(message)s",
        "%Y-%m-%d %H:%M:%S",
    )

    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(logging.DEBUG)
    console_handler.setFormatter(default_formatter)
    handlers = [console_handler]
    if log_file_path:
        file_handler = logging.FileHandler(log_file_path)
        file_handler.setFormatter(default_formatter)
        file_handler.setLevel(logging.DEBUG)
        handlers.append(file_handler)

    return handlers


def configure_logging(log_file_path: Optional[str] = ""):
    if len(logging.getLogger().handlers) == 0:
        handlers = get_logging_handlers(log_file_path)
        logging.getLogger("").handlers.extend(handlers)
        logging.root.setLevel(logging.DEBUG)
        logger = logging.getLogger(__name__)
        logger.info(f"Added logging handlers: {handlers}")
    else:
        logger = logging.getLogger(__name__)
        logger.info("Logging already configured!")



import numpy as np
from ultralytics import SAM

from skellytracker.trackers.base_tracker.base_tracker import BaseTracker


class SAMTracker(BaseTracker):
    def __init__(self):
        super().__init__(recorder=None, tracked_object_names=["segmentation"])

        self.model = SAM("sam_b.pt")

    def process_image(self, image, **kwargs):
        results = self.model.predict(image)

        self.tracked_objects["segmentation"].extra["landmarks"] = np.array(
            results[0].keypoints
        )

        self.annotated_image = self.annotate_image(image, results=results, **kwargs)

        return self.tracked_objects

    def annotate_image(self, image: np.ndarray, results, **kwargs) -> np.ndarray:
        return results[0].plot()


if __name__ == "__main__":
    SAMTracker().demo()



class MediapipeBlendshapeModelInfo:
    name = "mediapipe_blendshape"
    tracker_name = "MediapipeBlendshapeTracker"
    landmark_names = [
        "_neutral",
        "browDownLeft",
        "browDownRight",
        "browInnerUp",
        "browOuterUpLeft",
        "browOuterUpRight",
        "cheekPuff",
        "cheekSquintLeft",
        "cheekSquintRight",
        "eyeBlinkLeft",
        "eyeBlinkRight",
        "eyeLookDownLeft",
        "eyeLookDownRight",
        "eyeLookInLeft",
        "eyeLookInRight",
        "eyeLookOutLeft",
        "eyeLookOutRight",
        "eyeLookUpLeft",
        "eyeLookUpRight",
        "eyeSquintLeft",
        "eyeSquintRight",
        "eyeWideLeft",
        "eyeWideRight",
        "jawForward",
        "jawLeft",
        "jawOpen",
        "jawRight",
        "mouthClose",
        "mouthDimpleLeft",
        "mouthDimpleRight",
        "mouthFrownLeft",
        "mouthFrownRight",
        "mouthFunnel",
        "mouthLeft",
        "mouthLowerDownLeft",
        "mouthLowerDownRight",
        "mouthPressLeft",
        "mouthPressRight",
        "mouthPucker",
        "mouthRight",
        "mouthRollLower",
        "mouthRollUpper",
        "mouthShrugLower",
        "mouthShrugUpper",
        "mouthSmileLeft",
        "mouthSmileRight",
        "mouthStretchLeft",
        "mouthStretchRight",
        "mouthUpperUpLeft",
        "mouthUpperUpRight",
        "noseSneerLeft",
        "noseSneerRight",
    ]
    num_tracked_points = len(landmark_names)
    tracked_object_names = ["face"]



from abc import ABC, abstractmethod
import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
import cv2
import numpy as np
from tqdm import tqdm


from skellytracker.trackers.base_tracker.base_recorder import BaseCumulativeRecorder, BaseRecorder
from skellytracker.trackers.base_tracker.tracked_object import TrackedObject
from skellytracker.trackers.base_tracker.video_handler import VideoHandler
from skellytracker.trackers.demo_viewers.image_demo_viewer import ImageDemoViewer
from skellytracker.trackers.demo_viewers.webcam_demo_viewer import (
    WebcamDemoViewer,
)

logger = logging.getLogger(__name__)


class BaseTracker(ABC):
    """
    An abstract base class for implementing different tracking algorithms.
    """

    def __init__(
        self,
        recorder: BaseRecorder,
        tracked_object_names: List[str],
        **data: Any,
    ):
        self.recorder = recorder
        self.annotated_image = None
        self.tracked_objects: Dict[str, TrackedObject] = {}

        for name in tracked_object_names:
            self.tracked_objects[name] = TrackedObject(object_id=name)

    @abstractmethod
    def process_image(self, image: np.ndarray, **kwargs) -> Dict[str, TrackedObject]:
        """
        Process the input image and apply the tracking algorithm.

        :param image: An input image.
        :return: A dictionary of tracked objects
        """
        pass

    @abstractmethod
    def annotate_image(
        self, image: np.ndarray, tracked_objects: Dict[str, TrackedObject], **kwargs
    ) -> np.ndarray:
        """
        Annotate the input image with the results of the tracking algorithm.

        :param image: An input image.
        :param tracked_objects: A dictionary of tracked objects.
        :return: Annotated image
        """
        pass

    def process_video(
        self,
        input_video_filepath: Union[str, Path],
        output_video_filepath: Optional[Union[str, Path]] = None,
        save_data_bool: bool = False,
        use_tqdm: bool = True,
    ) -> Union[np.ndarray, None]:
        """
        Run the tracker on a video.

        :param input_video_filepath: Path to video file.
        :param output_video_filepath: Path to save annotated video to, does not save video if None.
        :param save_data_bool: Whether to save the data to a file.
        :param use_tqdm: Whether to use tqdm to show a progress bar
        :return: Array of tracked keypoint data if tracker has an associated recorder
        """

        cap = cv2.VideoCapture(str(input_video_filepath))

        width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        image_size = (width, height)

        fps = cap.get(cv2.CAP_PROP_FPS)

        if output_video_filepath is not None:
            video_handler = VideoHandler(
                output_path=output_video_filepath, frame_size=image_size, fps=fps
            )
        else:
            video_handler = None

        ret, frame = cap.read()

        number_of_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

        if use_tqdm:
            iterator = tqdm(
                range(number_of_frames),
                desc=f"processing video: {Path(input_video_filepath).name}",
                total=number_of_frames,
                colour="magenta",
                unit="frames",
                dynamic_ncols=True,
            )
        else:
            iterator = range(number_of_frames)

        for _frame_number in iterator:
            if not ret or frame is None:
                logger.error(
                    f"Failed to load an image from: {str(input_video_filepath)}"
                )
                raise ValueError("Failed to load an image from: " + str(input_video_filepath))

            self.process_image(frame)
            if self.recorder is not None:
                self.recorder.record(self.tracked_objects)
            if video_handler is not None:
                if self.annotated_image is None:
                    self.annotated_image = frame
                video_handler.add_frame(self.annotated_image)

            ret, frame = cap.read()

        cap.release()
        if video_handler is not None:
            video_handler.close()

        output_array = self.process_and_save_tracked_objects(
            input_video_filepath, save_data_bool, image_size
        )

        self.cleanup()

        return output_array

    def process_and_save_tracked_objects(
        self,
        input_video_filepath: Union[str, Path],
        save_data_bool: bool,
        image_size: tuple,
    ) -> Optional[np.ndarray]:
        if self.recorder is not None:
            output_array = self.recorder.process_tracked_objects(image_size=image_size)
            if save_data_bool:
                self.recorder.save(
                    file_path=str(Path(input_video_filepath).with_suffix(".npy"))
                )
        else:
            output_array = None
        return output_array

    def cleanup(self) -> None:
        """
        Run any cleanup code for the tracker, including clearing the recorded objects.

        Can be overridden by subclasses if any tracker needs a specific cleanup.
        """
        if self.recorder is not None:
            self.recorder.clear_recorded_objects()

    def demo(self) -> None:
        """
        Run a demo of the tracker.

        :param window_title: The title of the demo window.
        :return: None
        """
        camera_viewer = WebcamDemoViewer(
            tracker=self, recorder=self.recorder, window_title=self.__class__.__name__
        )
        camera_viewer.run()

    def image_demo(self, image_path: Path) -> None:
        """
        Run tracker on single image

        :return: None
        """

        image_viewer = ImageDemoViewer(self, self.__class__.__name__)
        image_viewer.run(image_path=image_path)


class BaseCumulativeTracker(BaseTracker):
    """
    A base class for tracking algorithms that run cumulatively, i.e are not able to process videos frame by frame.
    Throws a descriptive error for the abstract methods of BaseTracker that do not apply to this type of tracker.
    Trackers inheriting from this will need to overwrite the `process_video` method.
    """

    def __init__(
        self,
        tracked_object_names: List[str],
        recorder: BaseCumulativeRecorder,
        **data: Any,
    ):
        super().__init__(
            tracked_object_names=tracked_object_names, recorder=recorder, **data
        )

    def process_image(self, **kwargs) -> None:
        raise NotImplementedError(
            "This tracker does not support processing individual images, please use process_video instead."
        )

    def annotate_image(self, **kwargs) -> None:
        raise NotImplementedError(
            "This tracker does not support processing individual images, please use process_video instead."
        )

    @abstractmethod
    def process_video(
        self,
        input_video_filepath: Union[str, Path],
        output_video_filepath: Optional[Union[str, Path]] = None,
        save_data_bool: bool = False,
        use_tqdm: bool = True,
        **kwargs,
    ) -> Union[np.ndarray, None]:
        """
        Run the tracker on a video.

        :param input_video_filepath: Path to video file.
        :param output_video_filepath: Path to save annotated video to, does not save video if None.
        :param save_data_bool: Whether to save the data to a file.
        :param use_tqdm: Whether to use tqdm to show a progress bar
        :return: Array of tracked keypoint data
        """
        pass

    def image_demo(self, image_path: Path) -> None:
        raise NotImplementedError(
            "This tracker does not support processing individual images, please use process_video instead."
        )



from skellytracker.trackers.base_tracker.base_tracking_params import BaseTrackingParams


yolo_object_model_dictionary = {
    "nano": "yolov8n.pt",
    "small": "yolov8s.pt",
    "medium": "yolov8m.pt",
    "large": "yolov8l.pt",
    "extra_large": "yolov8x.pt",
}


class YOLOObjectTrackingParams(BaseTrackingParams):
    model_size: str = "medium"
    person_only: bool = True
    confidence_threshold: float = 0.5



from typing import List, Literal
from mediapipe.python.solutions import holistic as mp_holistic
from mediapipe.python.solutions.face_mesh import FACEMESH_NUM_LANDMARKS_WITH_IRISES

from skellytracker.trackers.base_tracker.base_tracking_params import BaseTrackingParams
from skellytracker.trackers.base_tracker.model_info import ModelInfo


# values for segment weight and segment mass percentages taken from Winter anthropometry tables
# https://imgur.com/a/aD74j
# Winter, D.A. (2005) Biomechanics and Motor Control of Human Movement. 3rd Edition, John Wiley & Sons, Inc., Hoboken.
class MediapipeModelInfo(ModelInfo):
    name = "mediapipe"
    tracker_name = "MediapipeHolisticTracker"
    body_landmark_names = [
        landmark.name.lower() for landmark in mp_holistic.PoseLandmark
    ]
    hand_landmark_names = [
        landmark.name.lower() for landmark in mp_holistic.HandLandmark
    ]
    face_landmark_names = [
        "right_eye",
        "left_eye",
        "nose_tip",
        "mouth_center",
        "right_ear_tragion",
        "left_ear_tragion",
    ]
    landmark_names = body_landmark_names + hand_landmark_names + face_landmark_names
    num_tracked_points_body = len(body_landmark_names)
    num_tracked_points_face = FACEMESH_NUM_LANDMARKS_WITH_IRISES
    num_tracked_points_left_hand = len(hand_landmark_names)
    num_tracked_points_right_hand = len(hand_landmark_names)
    num_tracked_points = (
        len(body_landmark_names)
        + 2 * len(hand_landmark_names)
        + num_tracked_points_face
    )
    tracked_object_names = [
        "pose_landmarks",
        "right_hand_landmarks",
        "left_hand_landmarks",
        "face_landmarks",
    ]
    virtual_markers_definitions = {
        "head_center": {
            "marker_names": ["left_ear", "right_ear"],
            "marker_weights": [0.5, 0.5],
        },
        "neck_center": {
            "marker_names": ["left_shoulder", "right_shoulder"],
            "marker_weights": [0.5, 0.5],
        },
        "trunk_center": {
            "marker_names": [
                "left_shoulder",
                "right_shoulder",
                "left_hip",
                "right_hip",
            ],
            "marker_weights": [0.25, 0.25, 0.25, 0.25],
        },
        "hips_center": {
            "marker_names": ["left_hip", "right_hip"],
            "marker_weights": [0.5, 0.5],
        },
    }
    segment_connections = {
        "head": {"proximal": "left_ear", "distal": "right_ear"},
        "neck": {
            "proximal": "head_center",
            "distal": "neck_center",
        },
        "spine": {
            "proximal": "neck_center",
            "distal": "hips_center",
        },
        "right_shoulder": {"proximal": "neck_center", "distal": "right_shoulder"},
        "left_shoulder": {"proximal": "neck_center", "distal": "left_shoulder"},
        "right_upper_arm": {"proximal": "right_shoulder", "distal": "right_elbow"},
        "left_upper_arm": {"proximal": "left_shoulder", "distal": "left_elbow"},
        "right_forearm": {"proximal": "right_elbow", "distal": "right_wrist"},
        "left_forearm": {"proximal": "left_elbow", "distal": "left_wrist"},
        "right_hand": {"proximal": "right_wrist", "distal": "right_index"},
        "left_hand": {"proximal": "left_wrist", "distal": "left_index"},
        "right_pelvis": {"proximal": "hips_center", "distal": "right_hip"},
        "left_pelvis": {"proximal": "hips_center", "distal": "left_hip"},
        "right_thigh": {"proximal": "right_hip", "distal": "right_knee"},
        "left_thigh": {"proximal": "left_hip", "distal": "left_knee"},
        "right_shank": {"proximal": "right_knee", "distal": "right_ankle"},
        "left_shank": {"proximal": "left_knee", "distal": "left_ankle"},
        "right_foot": {"proximal": "right_ankle", "distal": "right_foot_index"},
        "left_foot": {"proximal": "left_ankle", "distal": "left_foot_index"},
        "right_heel": {"proximal": "right_ankle", "distal": "right_heel"},
        "left_heel": {"proximal": "left_ankle", "distal": "left_heel"},
        "right_foot_bottom": {"proximal": "right_heel", "distal": "right_foot_index"},
        "left_foot_bottom": {"proximal": "left_heel", "distal": "left_foot_index"},
    }
    center_of_mass_definitions = {
        "head": {
            "segment_com_length": 0.5,
            "segment_com_percentage": 0.081,
        },
        "spine": {
            "segment_com_length": 0.5,
            "segment_com_percentage": 0.497,
        },
        "right_upper_arm": {
            "segment_com_length": 0.436,
            "segment_com_percentage": 0.028,
        },
        "left_upper_arm": {
            "segment_com_length": 0.436,
            "segment_com_percentage": 0.028,
        },
        "right_forearm": {
            "segment_com_length": 0.430,
            "segment_com_percentage": 0.016,
        },
        "left_forearm": {
            "segment_com_length": 0.430,
            "segment_com_percentage": 0.016,
        },
        "right_hand": {
            "segment_com_length": 0.506,
            "segment_com_percentage": 0.006,
        },
        "left_hand": {
            "segment_com_length": 0.506,
            "segment_com_percentage": 0.006,
        },
        "right_thigh": {
            "segment_com_length": 0.433,
            "segment_com_percentage": 0.1,
        },
        "left_thigh": {
            "segment_com_length": 0.433,
            "segment_com_percentage": 0.1,
        },
        "right_shank": {
            "segment_com_length": 0.433,
            "segment_com_percentage": 0.0465,
        },
        "left_shank": {
            "segment_com_length": 0.433,
            "segment_com_percentage": 0.0465,
        },
        "right_foot": {
            "segment_com_length": 0.5,
            "segment_com_percentage": 0.0145,
        },
        "left_foot": {
            "segment_com_length": 0.5,
            "segment_com_percentage": 0.0145,
        },
    }
    joint_hierarchy = {
        "hips_center": ["left_hip", "right_hip", "trunk_center"],
        "trunk_center": ["neck_center"],
        "neck_center": ["left_shoulder", "right_shoulder", "head_center"],
        "head_center": [
            "nose",
            "left_eye_inner",
            "left_eye",
            "left_eye_outer",
            "right_eye_inner",
            "right_eye",
            "right_eye_outer",
            "left_ear",
            "right_ear",
            "mouth_left",
            "mouth_right",
        ],
        "left_shoulder": ["left_elbow"],
        "left_elbow": ["left_wrist"],
        "left_wrist": ["left_pinky", "left_index", "left_thumb"],
        "right_shoulder": ["right_elbow"],
        "right_elbow": ["right_wrist"],
        "right_wrist": ["right_pinky", "right_index", "right_thumb"],
        "left_hip": ["left_knee"],
        "left_knee": ["left_ankle"],
        "left_ankle": ["left_heel", "left_foot_index"],
        "right_hip": ["right_knee"],
        "right_knee": ["right_ankle"],
        "right_ankle": ["right_heel", "right_foot_index"],
    }


class MediapipeTrackingParams(BaseTrackingParams):
    use_yolo_crop_method: bool = False
    mediapipe_model_complexity: int = 2
    min_detection_confidence: float = 0.5
    min_tracking_confidence: float = 0.5
    static_image_mode: bool = False
    yolo_model_size: Literal[
        "nano", "small", "medium", "large", "extra_large", "high_res"
    ] = "nano"
    bounding_box_buffer_percentage: float = 10
    buffer_size_method: Literal["buffer_by_box_size", "buffer_by_image_size"] = (
        "buffer_by_box_size"
    )


def mediapipe_body_names_match_expected(
    mediapipe_body_landmark_names: List[str],
) -> bool:
    """
    Check if the mediapipe folks have changed their landmark names. If they have, then this function may need to be updated.

    Args:
        mediapipe_body_landmark_names: List of strings, each string is the name of a mediapipe landmark.

    Returns:
        bool: True if the mediapipe landmark names are as expected, False otherwise.
    """
    expected_mediapipe_body_landmark_names = [
        "nose",
        "left_eye_inner",
        "left_eye",
        "left_eye_outer",
        "right_eye_inner",
        "right_eye",
        "right_eye_outer",
        "left_ear",
        "right_ear",
        "mouth_left",
        "mouth_right",
        "left_shoulder",
        "right_shoulder",
        "left_elbow",
        "right_elbow",
        "left_wrist",
        "right_wrist",
        "left_pinky",
        "right_pinky",
        "left_index",
        "right_index",
        "left_thumb",
        "right_thumb",
        "left_hip",
        "right_hip",
        "left_knee",
        "right_knee",
        "left_ankle",
        "right_ankle",
        "left_heel",
        "right_heel",
        "left_foot_index",
        "right_foot_index",
    ]
    return mediapipe_body_landmark_names == expected_mediapipe_body_landmark_names



from copy import deepcopy
from typing import Dict
import numpy as np

from skellytracker.trackers.base_tracker.base_recorder import BaseRecorder
from skellytracker.trackers.base_tracker.tracked_object import TrackedObject
from skellytracker.trackers.yolo_tracker.yolo_model_info import YOLOModelInfo


class YOLORecorder(BaseRecorder):
    def record(self, tracked_objects: Dict[str, TrackedObject]) -> None:
        self.recorded_objects.append(deepcopy(tracked_objects["tracked_person"]))

    def process_tracked_objects(self, **kwargs) -> np.ndarray:
        self.recorded_objects_array = np.zeros(
            (len(self.recorded_objects), YOLOModelInfo.num_tracked_points, 3)
        )

        for i, recorded_object in enumerate(self.recorded_objects):
            for j in range(YOLOModelInfo.num_tracked_points):
                self.recorded_objects_array[i, j, 0] = recorded_object.extra[
                    "landmarks"
                ][0, j, 0]
                self.recorded_objects_array[i, j, 1] = recorded_object.extra[
                    "landmarks"
                ][0, j, 1]
                self.recorded_objects_array[i, j, 2] = np.nan

        return self.recorded_objects_array



import logging
from typing import Dict

import cv2
import numpy as np
from pydantic import BaseModel

from skellytracker.trackers.base_tracker.base_tracker import BaseTracker
from skellytracker.trackers.base_tracker.tracked_object import TrackedObject
from skellytracker.trackers.bright_point_tracker.brightest_point_recorder import (
    BrightestPointRecorder,
)

UPPER_BOUND_COLOR = [255, 255, 255]

LOWER_BOUND_COLOR = [40, 40, 40]

logger = logging.getLogger(__name__)


class BrightPatch(BaseModel):
    area: float
    centroid_x: int
    centroid_y: int


class BrightestPointTracker(BaseTracker):
    def __init__(self, num_points: int = 1, luminance_threshold: int = 200):
        super().__init__(
            tracked_object_names=[f"brightest_point_{i}" for i in range(num_points)],
            recorder=BrightestPointRecorder(),
        )

        self.num_points = num_points
        self.luminance_threshold = luminance_threshold

    def process_image(self, image: np.ndarray, **kwargs) -> Dict[str, TrackedObject]:
        # Convert the image to grayscale
        gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

        # Threshold the image to get only bright regions
        _, thresholded_image = cv2.threshold(
            gray_image, self.luminance_threshold, 255, cv2.THRESH_BINARY
        )

        # Find contours of the bright regions
        bright_patches, _ = cv2.findContours(
            thresholded_image, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
        )

        # Process each bright patch separately
        patch_list = []
        for patch in bright_patches:
            patch_moments = cv2.moments(patch)
            if patch_moments["m00"] != 0:  # Avoid division by zero
                centroid_x = int(patch_moments["m10"] / patch_moments["m00"])
                centroid_y = int(patch_moments["m01"] / patch_moments["m00"])

                patch_list.append(
                    BrightPatch(
                        area=cv2.contourArea(patch),
                        centroid_x=centroid_x,
                        centroid_y=centroid_y,
                    )
                )

        largest_patches = sorted(
            patch_list, key=lambda patch: patch.area, reverse=True
        )[: self.num_points]

        for i, patch in enumerate(largest_patches):
            self.tracked_objects[f"brightest_point_{i}"].pixel_x = patch.centroid_x
            self.tracked_objects[f"brightest_point_{i}"].pixel_y = patch.centroid_y
            self.tracked_objects[f"brightest_point_{i}"].extra[
                "thresholdedimage"
            ] = thresholded_image

        for i in range(len(largest_patches), self.num_points):
            self.tracked_objects[f"brightest_point_{i}"].pixel_x = (
                None  # TODO: Is this the right value for missing data?
            )
            self.tracked_objects[f"brightest_point_{i}"].pixel_y = None

        self.annotated_image = self.annotate_image(
            image=image, tracked_objects=self.tracked_objects
        )

        return self.tracked_objects

    def annotate_image(
        self, image: np.ndarray, tracked_objects: Dict[str, TrackedObject], **kwargs
    ) -> np.ndarray:
        annotated_image = image.copy()

        for key, tracked_object in tracked_objects.items():
            if (
                "brightest_point" in key
                and tracked_object.pixel_x is not None
                and tracked_object.pixel_y is not None
            ):
                cv2.drawMarker(
                    img=annotated_image,
                    position=(int(tracked_object.pixel_x), int(tracked_object.pixel_y)),
                    color=(0, 0, 255),
                    markerType=cv2.MARKER_CROSS,
                    markerSize=20,
                    thickness=2,
                )

        return annotated_image


if __name__ == "__main__":
    BrightestPointTracker(num_points=2).demo()



from abc import ABC, abstractmethod
import logging
from pathlib import Path
from typing import Dict, Union, Optional

import numpy as np

from skellytracker.trackers.base_tracker.tracked_object import TrackedObject

logger = logging.getLogger(__name__)


class BaseRecorder(ABC):
    """
    An abstract base class for implementing different recording algorithms.
    """

    def __init__(self):
        self.recorded_objects = []
        self.recorded_objects_array = None

    @abstractmethod
    def record(
        self, tracked_objects: Dict[str, TrackedObject], annotated_image: Optional[np.ndarray] = None
    ) -> None:
        """
        Record the tracked objects as they are created by the tracker.

        :param tracked_object: A tracked objects dictionary.
        :return: None
        """
        pass

    @abstractmethod
    def process_tracked_objects(self, **kwargs) -> np.ndarray:
        """
        Process the recorded objects to be in a consistent array format across trackers.

        :return: Array of tracked objects.
        """
        pass

    def clear_recorded_objects(self):
        logger.info("Clearing recorded objects from recorder")
        self.recorded_objects = []
        self.recorded_objects_array = None

    def save(self, file_path: Union[str, Path]) -> None:
        """
        Save the recorded objects to a file.

        :param file_path: The path to the file where the recorded objects should be saved.
        :return: None
        """
        if self.recorded_objects_array is None:
            recorded_objects_array = self.process_tracked_objects()
        else:
            recorded_objects_array = self.recorded_objects_array
        logger.info(f"Saving recorded objects to {file_path}")
        np.save(file_path, recorded_objects_array)


class BaseCumulativeRecorder(BaseRecorder):
    """
    A base class for recording data from cumulative trackers.
    Throws a descriptive error for methods that do not apply to recording data from this type of tracker.
    Trackers implementing this will only use the process_tracked_objects method to get data in the proper format.
    """

    def __init__(self):
        super().__init__()

    def record(self, tracked_objects: Dict[str, TrackedObject]) -> None:
        raise NotImplementedError(
            "This tracker does not support by frame recording, please use process_tracked_objects instead"
        )



import json
from typing import Dict, Union
import numpy as np
from pathlib import Path
import re
from tqdm import tqdm

from skellytracker.trackers.base_tracker.base_recorder import BaseCumulativeRecorder
from skellytracker.trackers.openpose_tracker.openpose_model_info import (
    OpenPoseModelInfo,
)


class OpenPoseRecorder(BaseCumulativeRecorder):
    def __init__(
        self,
        track_hands: bool = False,
        track_faces: bool = False,
    ):
        super().__init__()
        self.track_hands = track_hands
        self.track_faces = track_faces

    def extract_frame_index(self, filename: str) -> Union[int, None]:
        """Extract the numeric part indicating the frame index from the filename."""
        match = re.search(r"_(\d{12})_keypoints", filename)
        return int(match.group(1)) if match else None

    def parse_openpose_jsons(self, json_directory: Union[Path, str]) -> np.ndarray:
        # Remove the iteration over subdirectories and focus on a single directory
        json_directory = Path(json_directory)
        files = list(Path(json_directory).glob("*.json"))
        num_frames = len(files)
        frame_indices = [
            index
            for f in files
            if (index := self.extract_frame_index(f.name)) is not None
        ]
        frame_indices.sort()

        if len(frame_indices) != num_frames:
            raise ValueError(
                f"Invalid number of frames in {json_directory}: expected {num_frames} != {len(frame_indices)} frames in file"
            )

        num_markers = OpenPoseModelInfo.num_tracked_points_body
        if self.track_hands:
            num_markers += (
                OpenPoseModelInfo.num_tracked_points_right_hand
                + OpenPoseModelInfo.num_tracked_points_left_hand
            )
        if self.track_faces:
            num_markers += OpenPoseModelInfo.num_tracked_points_face

        # Initialize a single camera array since we're only processing one video at a time
        data_array = np.full((num_frames, num_markers, 3), np.nan)

        # Process each JSON file in the directory
        for file_index, json_file in enumerate(
            tqdm(files, desc=f"Processing {json_directory.name} JSONs")
        ):
            with open(json_file) as f:
                data = json.load(f)

            if data["people"]:
                keypoints = self.extract_keypoints(data["people"][0])
                data_array[frame_indices[file_index], :, :] = keypoints

        return data_array

    def extract_keypoints(self, person_data: Dict[str, np.ndarray]) -> np.ndarray:
        """Extract and organize keypoints from person data."""

        body_markers = OpenPoseModelInfo.num_tracked_points_body
        hand_markers = (
            OpenPoseModelInfo.num_tracked_points_left_hand
            + OpenPoseModelInfo.num_tracked_points_right_hand
        )
        face_markers = OpenPoseModelInfo.num_tracked_points_face

        # Initialize a full array of NaNs for keypoints
        keypoints_array = np.full(
            (body_markers + (hand_markers) + face_markers, 3), np.nan
        )

        # Populate the array with available data
        if "pose_keypoints_2d" in person_data:
            keypoints_array[:body_markers, :] = np.reshape(
                person_data["pose_keypoints_2d"], (-1, 3)
            )[:body_markers, :]
        if (
            "hand_left_keypoints_2d" in person_data
            and "hand_right_keypoints_2d" in person_data
        ):
            keypoints_array[body_markers : body_markers + OpenPoseModelInfo.num_tracked_points_left_hand, :] = np.reshape(
                person_data["hand_left_keypoints_2d"], (-1, 3)
            )[:hand_markers, :]
            keypoints_array[
                body_markers + OpenPoseModelInfo.num_tracked_points_left_hand : body_markers + OpenPoseModelInfo.num_tracked_points_left_hand + OpenPoseModelInfo.num_tracked_points_right_hand, :
            ] = np.reshape(person_data["hand_right_keypoints_2d"], (-1, 3))[
                :hand_markers, :
            ]
        if "face_keypoints_2d" in person_data:
            keypoints_array[body_markers + hand_markers :, :] = np.reshape(
                person_data["face_keypoints_2d"], (-1, 3)
            )[:face_markers, :]

        return keypoints_array

    def process_tracked_objects(self, output_json_path: Path) -> np.ndarray:
        """
        Convert the recorded JSON data into the structured numpy array format.
        """
        # In this case, the recorded_objects are already in the desired format, so we simply return them.
        self.recorded_objects_array = self.parse_openpose_jsons(output_json_path)
        return self.recorded_objects_array



# __main__.py
import sys
from pathlib import Path

base_package_path = Path(__file__).parent.parent
print(f"adding base_package_path: {base_package_path} : to sys.path")
sys.path.insert(0, str(base_package_path))  # add parent directory to sys.path

import logging  # noqa: E402

from skellytracker.RUN_ME import main  # noqa: E402

logger = logging.getLogger(__name__)


def cli_main():
    logger.info("Running as a script")
    if len(sys.argv) > 1:
        demo_tracker = str(sys.argv[1])
    else:
        demo_tracker = "mediapipe_holistic_tracker"
    main(demo_tracker=demo_tracker)


if __name__ == "__main__":
    cli_main()



import cv2
import pytest
import numpy as np


from skellytracker.trackers.bright_point_tracker.brightest_point_tracker import (
    BrightestPointTracker,
)


@pytest.fixture()
def sample_image():
    """
    Create a sample image with bright spots for testing.
    """
    image = np.zeros((100, 100, 3), dtype=np.uint8)
    cv2.circle(image, (30, 30), 10, (255, 255, 255), -1)  # Bright spot
    cv2.circle(image, (70, 70), 15, (255, 255, 255), -1)  # Brighter spot
    return image


def test_process_image_with_one_brightest_point(sample_image):
    tracker = BrightestPointTracker(num_points=1, luminance_threshold=200)
    tracked_objects = tracker.process_image(sample_image)

    assert len(tracked_objects) == 1
    assert tracked_objects["brightest_point_0"].pixel_x == 70
    assert tracked_objects["brightest_point_0"].pixel_y == 70


def test_process_image_with_two_brightest_points(sample_image):
    tracker = BrightestPointTracker(num_points=2, luminance_threshold=200)
    tracked_objects = tracker.process_image(sample_image)

    assert len(tracked_objects) == 2
    assert tracked_objects["brightest_point_0"].pixel_x == 70
    assert tracked_objects["brightest_point_0"].pixel_y == 70
    assert tracked_objects["brightest_point_1"].pixel_x == 30
    assert tracked_objects["brightest_point_1"].pixel_y == 30


def test_process_image_with_five_brightest_points(sample_image):
    tracker = BrightestPointTracker(num_points=5, luminance_threshold=200)
    tracked_objects = tracker.process_image(sample_image)

    assert len(tracked_objects) == 5
    assert tracked_objects["brightest_point_0"].pixel_x == 70
    assert tracked_objects["brightest_point_0"].pixel_y == 70
    assert tracked_objects["brightest_point_1"].pixel_x == 30
    assert tracked_objects["brightest_point_1"].pixel_y == 30
    assert tracked_objects["brightest_point_2"].pixel_x is None
    assert tracked_objects["brightest_point_2"].pixel_y is None
    assert tracked_objects["brightest_point_3"].pixel_x is None
    assert tracked_objects["brightest_point_3"].pixel_y is None
    assert tracked_objects["brightest_point_4"].pixel_x is None
    assert tracked_objects["brightest_point_4"].pixel_y is None


def test_annotate_image(sample_image):
    tracker = BrightestPointTracker(num_points=2, luminance_threshold=200)
    tracked_objects = tracker.process_image(sample_image)

    assert tracker.annotated_image is not None

    # Check if the bright spots have markers
    bright_point_0 = tracked_objects["brightest_point_0"]
    bright_point_1 = tracked_objects["brightest_point_1"]

    assert bright_point_0.pixel_x is not None
    assert bright_point_0.pixel_y is not None
    assert bright_point_1.pixel_x is not None
    assert bright_point_1.pixel_y is not None

    assert tracker.annotated_image[
        int(bright_point_0.pixel_y), int(bright_point_0.pixel_x)
    ].tolist() == [0, 0, 255]
    assert tracker.annotated_image[
        int(bright_point_1.pixel_y), int(bright_point_1.pixel_x)
    ].tolist() == [0, 0, 255]


def test_record(sample_image):
    tracker = BrightestPointTracker(num_points=2, luminance_threshold=200)
    tracked_objects = tracker.process_image(sample_image)
    tracker.recorder.record(tracked_objects=tracked_objects)

    assert len(tracker.recorder.recorded_objects) == 1

    assert len(tracker.recorder.recorded_objects[0]) == 2



from mmpose.apis import inference_top_down_pose_model, init_pose_model
from skellytracker.trackers.base_tracker.base_tracker import BaseTracker

# correct/fill this out based on these docs: https://github.com/open-mmlab/mmpose/blob/main/docs/en/user_guides/inference.md


class MMPoseTracker(BaseTracker):
    def __init__(self, config_file, checkpoint_file):
        super().__init__(recorder=None, tracked_object_names=["human_pose"])
        self.model = init_pose_model(config_file, checkpoint_file, device="cuda:0")

    def process_image(self, image, **kwargs):
        pose_results, returned_outputs = inference_top_down_pose_model(
            self.model,
            image,
            bbox_thr=None,
            format="xyxy",
            dataset="TopDownCocoDataset",
            return_heatmap=False,
            outputs=None,
        )

        # Draw the poses on the image
        self.model.show_result(image, pose_results, show=False, out_file=None)

        # Update the tracking data
        self.tracking_data = {"human_pose": pose_results}
        self.annotated_image = image

        return {
            "tracking_data": self.tracking_data,
            "annotated_image": self.annotated_image,
            "raw_image": image,
        }


if __name__ == "__main__":
    config_file = "configs/top_down/hrnet/coco/hrnet_w48_coco_256x192.py"
    checkpoint_file = "https://download.openmmlab.com/mmpose/top_down/hrnet/hrnet_w48_coco_256x192-b9e0b3ab_20200708.pth"
    MMPoseTracker(config_file, checkpoint_file).demo()



from skellytracker.trackers.base_tracker.base_tracking_params import BaseTrackingParams
from skellytracker.trackers.base_tracker.model_info import ModelInfo


class YOLOModelInfo(ModelInfo):
    name = "yolo"
    tracker_name = "YOLOPoseTracker"
    num_tracked_points = 17
    model_dictionary = {
        "nano": "yolov8n-pose.pt",
        "small": "yolov8s-pose.pt",
        "medium": "yolov8m-pose.pt",
        "large": "yolov8l-pose.pt",
        "extra_large": "yolov8x-pose.pt",
        "high_res": "yolov8x-pose-p6.pt",
    } # TODO: rename to tracker_dictionary to avoid pydantic 2 conflict?
    landmark_names = [
        "nose",
        "left_eye",
        "right_eye",
        "left_ear",
        "right_ear",
        "left_shoulder",
        "right_shoulder",
        "left_elbow",
        "right_elbow",
        "left_wrist",
        "right_wrist",
        "left_hip",
        "right_hip",
        "left_knee",
        "right_knee",
        "left_ankle",
        "right_ankle",
    ]
    virtual_markers_definitions = {
        "head_center": {
            "marker_names": ["left_ear", "right_ear"],
            "marker_weights": [0.5, 0.5],
        },
        "neck_center": {
            "marker_names": ["left_shoulder", "right_shoulder"],
            "marker_weights": [0.5, 0.5],
        },
        "trunk_center": {
            "marker_names": [
                "left_shoulder",
                "right_shoulder",
                "left_hip",
                "right_hip",
            ],
            "marker_weights": [0.25, 0.25, 0.25, 0.25],
        },
        "hips_center": {
            "marker_names": ["left_hip", "right_hip"],
            "marker_weights": [0.5, 0.5],
        },
    }
    segment_connections = {
        "head": {"proximal": "left_ear", "distal": "right_ear"},
        "neck": {
            "proximal": "head_center",
            "distal": "neck_center",
        },
        "spine": {
            "proximal": "neck_center",
            "distal": "hips_center",
        },
        "right_shoulder": {"proximal": "neck_center", "distal": "right_shoulder"},
        "left_shoulder": {"proximal": "neck_center", "distal": "left_shoulder"},
        "right_upper_arm": {"proximal": "right_shoulder", "distal": "right_elbow"},
        "left_upper_arm": {"proximal": "left_shoulder", "distal": "left_elbow"},
        "right_forearm": {"proximal": "right_elbow", "distal": "right_wrist"},
        "left_forearm": {"proximal": "left_elbow", "distal": "left_wrist"},
        "right_pelvis": {"proximal": "hips_center", "distal": "right_hip"},
        "left_pelvis": {"proximal": "hips_center", "distal": "left_hip"},
        "right_thigh": {"proximal": "right_hip", "distal": "right_knee"},
        "left_thigh": {"proximal": "left_hip", "distal": "left_knee"},
        "right_shank": {"proximal": "right_knee", "distal": "right_ankle"},
        "left_shank": {"proximal": "left_knee", "distal": "left_ankle"},
    }
    center_of_mass_definitions = {
        "head": {
            "segment_com_length": 0.5,
            "segment_com_percentage": 0.081,
        },
        "spine": {
            "segment_com_length": 0.5,
            "segment_com_percentage": 0.497,
        },
        "right_upper_arm": {
            "segment_com_length": 0.436,
            "segment_com_percentage": 0.028,
        },
        "left_upper_arm": {
            "segment_com_length": 0.436,
            "segment_com_percentage": 0.028,
        },
        "right_forearm": {
            "segment_com_length": 0.682,
            "segment_com_percentage": 0.022,
        },
        "left_forearm": {
            "segment_com_length": 0.682,
            "segment_com_percentage": 0.022,
        },
        "right_thigh": {
            "segment_com_length": 0.433,
            "segment_com_percentage": 0.1,
        },
        "left_thigh": {
            "segment_com_length": 0.433,
            "segment_com_percentage": 0.1,
        },
        "right_shank": {
            "segment_com_length": 0.606,
            "segment_com_percentage": 0.061,
        },
        "left_shank": {
            "segment_com_length": 0.606,
            "segment_com_percentage": 0.061,
        },
    }
    joint_hierarchy = {
        "hips_center": ["left_hip", "right_hip", "trunk_center"],
        "trunk_center": ["neck_center"],
        "neck_center": ["left_shoulder", "right_shoulder", "head_center"],
        "head_center": [
            "nose",
            "left_eye",
            "right_eye",
            "left_ear",
            "right_ear",
        ],
        "left_shoulder": ["left_elbow"],
        "left_elbow": ["left_wrist"],
        "right_shoulder": ["right_elbow"],
        "right_elbow": ["right_wrist"],
        "left_hip": ["left_knee"],
        "left_knee": ["left_ankle"],
        "right_hip": ["right_knee"],
        "right_knee": ["right_ankle"],
    }


class YOLOTrackingParams(BaseTrackingParams):
    model_size: str = "medium" # TODO: rename to tracker_model_size to avoid pydantic 2 conflict?



import cv2
from pathlib import Path

from skellytracker.trackers.bright_point_tracker.brightest_point_tracker import (
    BrightestPointTracker,
)
from skellytracker.trackers.charuco_tracker.charuco_tracker import CharucoTracker

try:
    from skellytracker.trackers.mediapipe_tracker.mediapipe_holistic_tracker import (
        MediapipeHolisticTracker,
    )
except ModuleNotFoundError:
    print("\n\nTo use mediapipe_holistic_tracker, install skellytracker[mediapipe]\n\n")
try:
    from skellytracker.trackers.yolo_tracker.yolo_tracker import YOLOPoseTracker
except ModuleNotFoundError:
    print("\n\nTo use yolo_tracker, install skellytracker[yolo]\n\n")


if __name__ == "__main__":
    demo_tracker = "brightest_point_tracker"
    image_path = Path("/Path/To/Your/Image.jpg")

    if demo_tracker == "brightest_point_tracker":
        BrightestPointTracker(num_points=2).image_demo(image_path=image_path)

    elif demo_tracker == "charuco_tracker":
        charuco_squares_x = 7
        charuco_squares_y = 5
        number_of_charuco_markers = (charuco_squares_x - 1) * (charuco_squares_y - 1)
        charuco_ids = [str(index) for index in range(number_of_charuco_markers)]

        CharucoTracker(
            tracked_object_names=charuco_ids,
            squares_x=charuco_squares_x,
            squares_y=charuco_squares_y,
            dictionary=cv2.aruco.getPredefinedDictionary(cv2.aruco.DICT_4X4_250),
        ).image_demo(image_path=image_path)

    elif demo_tracker == "mediapipe_holistic_tracker":
        MediapipeHolisticTracker(
            model_complexity=2,
            min_detection_confidence=0.5,
            min_tracking_confidence=0.5,
            static_image_mode=False,
            smooth_landmarks=True,
        ).image_demo(image_path=image_path)

    elif demo_tracker == "yolo_tracker":
        YOLOPoseTracker(model_size="high_res").image_demo(image_path=image_path)



from copy import deepcopy
from typing import Dict
import numpy as np

from skellytracker.trackers.base_tracker.base_recorder import BaseRecorder
from skellytracker.trackers.base_tracker.tracked_object import TrackedObject


class CharucoRecorder(BaseRecorder):
    def record(self, tracked_objects: Dict[str, TrackedObject]) -> None:
        self.recorded_objects.append(
            [deepcopy(tracked_object) for tracked_object in tracked_objects.values()]
        )

    def process_tracked_objects(self, **kwargs) -> np.ndarray:
        self.recorded_objects_array = np.array(
            [
                [[tracked_object.pixel_x, tracked_object.pixel_y] for tracked_object in tracked_object_list]
                for tracked_object_list in self.recorded_objects
            ]
        )

        return self.recorded_objects_array



from copy import deepcopy
from typing import Dict
import numpy as np

from skellytracker.trackers.base_tracker.base_recorder import BaseRecorder
from skellytracker.trackers.base_tracker.tracked_object import TrackedObject


class YOLOObjectRecorder(BaseRecorder):
    def record(self, tracked_objects: Dict[str, TrackedObject]) -> None:
        self.recorded_objects.append(deepcopy(tracked_objects["object"]))

    def process_tracked_objects(self, **kwargs) -> np.ndarray:
        self.recorded_objects_array = np.zeros((len(self.recorded_objects), 4))
        for i, recorded_object in enumerate(self.recorded_objects):
            self.recorded_objects_array[i, :] = recorded_object.extra["boxes_xyxy"]

        return self.recorded_objects_array



from pathlib import Path
from typing import Optional
import cv2

# Constants for key actions
KEY_QUIT = ord("q")


class ImageDemoViewer:
    def __init__(self, tracker, window_title: Optional[str] = None):
        """
        Initialize with a tracker and optional window title and default exposure.
        """
        self.tracker = tracker
        if window_title is None:
            window_title = f"{tracker.__class__.__name__}"
        self.window_title = window_title

    def run(self, image_path: Path):
        """
        Display input image
        """
        image = cv2.imread(str(image_path))

        self.tracker.process_image(image)

        annotated_image = self.tracker.annotated_image

        cv2.imshow(self.window_title, annotated_image)

        cv2.waitKey(0)

        cv2.destroyAllWindows()



from copy import deepcopy
from typing import Dict
import numpy as np

from skellytracker.trackers.base_tracker.base_recorder import BaseRecorder
from skellytracker.trackers.base_tracker.tracked_object import TrackedObject
from skellytracker.trackers.mediapipe_blendshape_tracker.mediapipe_blendshape_model_info import (
    MediapipeBlendshapeModelInfo,
)


class MediapipeBlendshapeRecorder(BaseRecorder):
    def record(self, tracked_objects: Dict[str, TrackedObject]) -> None:
        self.recorded_objects.append([deepcopy(tracked_objects["face"])])

    def process_tracked_objects(self, **kwargs) -> np.ndarray:
        image_size = kwargs.get("image_size")
        if image_size is None:
            raise ValueError(
                f"image_size must be provided to process tracked objects from {__class__.__name__}"
            )
        self.recorded_objects_array = np.full(
            (
                len(self.recorded_objects),
                MediapipeBlendshapeModelInfo.num_tracked_points,
                1,
            ),
            np.nan,
        )

        for i, tracked_objects in enumerate(self.recorded_objects):
            if blendshapes := tracked_objects[0].extra.get("blendshapes"):
                self.recorded_objects_array[i, :, 0] = blendshapes

        return self.recorded_objects_array



import argparse
from pathlib import Path
from typing import Generator
import pandas as pd

from skellytracker.trackers.mediapipe_blendshape_tracker.mediapipe_blendshape_tracker import (
    MediapipeBlendshapeTracker,
)


def create_timestamp_generator() -> Generator[str, None, None]:
    """
    Create a generator that generates timestamps in the format "00:00:00:00.000", with leading values rolling over before 30
    """
    count = 1
    refresh_value = 30
    while True:
        parts = [(count // (refresh_value**i)) % refresh_value for i in range(4)]
        yield f"{parts[3]:02d}:{parts[2]:02d}:{parts[1]:02d}:{parts[0]:02d}.{count % 1000:03d}"
        count += 1


def blendshapes_to_csv(
    input_video_filepath: Path,
    output_video_filepath: Path,
    output_csv_filepath: Path,
):
    tracker = MediapipeBlendshapeTracker()
    output_array = tracker.process_video(
        input_video_filepath=Path(input_video_filepath),
        output_video_filepath=Path(output_video_filepath),
        save_data_bool=True,
    )
    if output_array is None:
        raise ValueError("output_array is None, aborting")

    output_array = output_array.reshape(
        output_array.shape[0], -1
    )  # ensure output array is 2D

    df = pd.DataFrame(
        output_array,
        columns=tracker.model_info.landmark_names,
    ).drop(columns=["_neutral"])

    timestamp_generator = create_timestamp_generator()
    df.insert(0, "Timestamp", [next(timestamp_generator) for _ in range(len(df))])

    df.insert(1, "BlendShapeCount", tracker.model_info.num_tracked_points)

    df.to_csv(Path(output_csv_filepath), index=False)


def main():
    parser = argparse.ArgumentParser(
        prog="skellytracker_blendshapes",
        description="Process a video file to extract MediaPipe blendshapes and save the output video and CSV data.",
        epilog="Thank you for using skellytracker!",
    )

    parser.add_argument(
        "-i", "--input", type=Path, help="Path to the input video file."
    )
    parser.add_argument(
        "-o", "--output-video", type=Path, help="Path to save the output video file to."
    )
    parser.add_argument(
        "-c", "--output-csv", type=Path, help="Path to save the output CSV file to."
    )

    args = parser.parse_args()

    input_path = args.input or Path(input("Enter the path to the video to be processed: "))

    output_video_path = args.output_video or Path(input(
        "Enter the path to save output video to: "
    ))

    output_csv_path = args.output_csv or Path(input(
        "Enter the path to save the output CSV file to: "
    ))

    if not input_path.exists():
        raise ValueError(f"Input video path does not exist: {input_path}")

    if output_video_path.suffix != ".mp4":
        print("Output video path must be a .mp4 file, changing extension to .mp4")
        output_video_path = output_video_path.with_suffix(".mp4")

    if output_csv_path.suffix != ".csv":
        print("Output CSV path must be a .csv file, changing extension to .csv")
        output_csv_path = output_csv_path.with_suffix(".csv")

    blendshapes_to_csv(input_path, output_video_path, output_csv_path)


if __name__ == "__main__":
    main()



import numpy as np
from typing import Dict
from ultralytics import YOLO

from skellytracker.trackers.base_tracker.base_tracker import BaseTracker
from skellytracker.trackers.base_tracker.tracked_object import TrackedObject
from skellytracker.trackers.yolo_object_tracker.yolo_object_model_info import (
    yolo_object_model_dictionary,
)
from skellytracker.trackers.yolo_object_tracker.yolo_object_recorder import (
    YOLOObjectRecorder,
)


class YOLOObjectTracker(BaseTracker):
    def __init__(
        self,
        model_size: str = "nano",
        person_only: bool = True,
        confidence_threshold: float = 0.5,
    ):
        super().__init__(tracked_object_names=["object"], recorder=YOLOObjectRecorder())

        pytorch_model = yolo_object_model_dictionary[model_size]
        self.model = YOLO(pytorch_model)
        self.confidence_threshold = confidence_threshold
        # TODO: When we expose this in freemocap, replace this with an int/list[int] to decide which class to track
        # TODO: Will also need to parameterize the "max_det" and setup tracker to take multiple tracked objects
        if person_only:
            self.classes = 0  # 0 is the YOLO class for person detection
        else:
            self.classes = None  # None includes all classes

    def process_image(self, image, **kwargs) -> Dict[str, TrackedObject]:
        results = self.model(
            image,
            classes=self.classes,
            max_det=1,
            verbose=False,
            conf=self.confidence_threshold,
        )

        box_xyxy = np.asarray(
            results[0].boxes.xyxy.cpu()
        ).flatten()  # On GPU, need to copy to CPU before np array conversion

        if box_xyxy.size > 0:
            self.tracked_objects["object"].pixel_x = (box_xyxy[0] + box_xyxy[2]) / 0.5
            self.tracked_objects["object"].pixel_y = (box_xyxy[1] + box_xyxy[3]) / 0.5

        self.tracked_objects["object"].extra["boxes_xyxy"] = box_xyxy
        self.tracked_objects["object"].extra["original_image_shape"] = results[
            0
        ].boxes.orig_shape

        self.annotated_image = self.annotate_image(image, results=results, **kwargs)

        return self.tracked_objects

    def annotate_image(self, image: np.ndarray, results, **kwargs) -> np.ndarray:
        return results[0].plot()


if __name__ == "__main__":
    YOLOObjectTracker().demo()



from typing import Dict, List

import cv2
import numpy as np

from skellytracker.trackers.base_tracker.base_tracker import BaseTracker
from skellytracker.trackers.base_tracker.tracked_object import TrackedObject
from skellytracker.trackers.charuco_tracker.charuco_recorder import CharucoRecorder


default_aruco_dictionary = cv2.aruco.getPredefinedDictionary(cv2.aruco.DICT_4X4_250)

class CharucoTracker(BaseTracker):
    def __init__(
        self,
        tracked_object_names: List[str],
        squares_x: int,
        squares_y: int,
        dictionary: cv2.aruco.Dictionary = default_aruco_dictionary,
        square_length: float = 1,
        marker_length: float = 0.8,
    ):
        super().__init__(
            recorder=CharucoRecorder(), tracked_object_names=tracked_object_names
        )
        self.board = cv2.aruco.CharucoBoard(
            size=(squares_x, squares_y),
            squareLength=square_length,
            markerLength=marker_length,
            dictionary=dictionary,
        )

        # Following most recent charuco detection documentation: https://docs.opencv.org/4.x/df/d4a/tutorial_charuco_detection.html
        self.charuco_detector = cv2.aruco.CharucoDetector(self.board)

        self.tracked_object_names = tracked_object_names
        self.dictionary = dictionary

    def process_image(self, image: np.ndarray, **kwargs) -> Dict[str, TrackedObject]:
        # Convert the image to grayscale
        gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

        charuco_corners, charuco_ids, _marker_corners, _marker_ids = (
            self.charuco_detector.detectBoard(gray_image)
        )

        self.reinitialize_tracked_objects()

        # If any Charuco corners were found
        if (
            charuco_corners is not None
            and charuco_ids is not None
            and len(charuco_corners) > 3
        ):
            # Create a TrackedObject for each corner
            for id, corner in zip(charuco_ids, charuco_corners):
                object_id = str(id).strip("[]")
                self.tracked_objects[object_id] = TrackedObject(object_id=object_id)
                self.tracked_objects[object_id].pixel_x = corner[0][0]
                self.tracked_objects[object_id].pixel_y = corner[0][1]

        self.annotated_image = self.annotate_image(
            image=image, tracked_objects=self.tracked_objects
        )

        return self.tracked_objects

    def annotate_image(
        self, image: np.ndarray, tracked_objects: Dict[str, TrackedObject], **kwargs
    ) -> np.ndarray:
        # Copy the original image for annotation
        annotated_image = image.copy()

        # Draw a marker for each tracked corner
        for tracked_object in tracked_objects.values():
            if (
                tracked_object.pixel_x is not None
                and tracked_object.pixel_y is not None
            ):
                cv2.drawMarker(
                    annotated_image,
                    (int(tracked_object.pixel_x), int(tracked_object.pixel_y)),
                    (0, 0, 255),
                    markerType=cv2.MARKER_CROSS,
                    markerSize=30,
                    thickness=2,
                )
                cv2.putText(
                    annotated_image,
                    tracked_object.object_id,
                    (int(tracked_object.pixel_x), int(tracked_object.pixel_y)),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    1,
                    (255, 0, 0),
                    2,
                )

        return annotated_image

    def reinitialize_tracked_objects(self) -> None:
        """
        Reinitialize tracked objects to clear previous frames data

        Unlike self.tracked_objects.clear(), this will ensure every tracked object has a value for each frame, even if its empty
        """
        for name in self.tracked_object_names:
            self.tracked_objects[name] = TrackedObject(object_id=name)


if __name__ == "__main__":
    charuco_squares_x_in = 7
    charuco_squares_y_in = 5
    number_of_charuco_markers = (charuco_squares_x_in - 1) * (charuco_squares_y_in - 1)
    charuco_ids = [str(index) for index in range(number_of_charuco_markers)]

    CharucoTracker(
        tracked_object_names=charuco_ids,
        squares_x=charuco_squares_x_in,
        squares_y=charuco_squares_y_in,
        dictionary=cv2.aruco.getPredefinedDictionary(cv2.aruco.DICT_4X4_250),
    ).demo()



import cv2
import requests
import mediapipe as mp
import numpy as np

from typing import Dict, Union
from pathlib import Path
from mediapipe.tasks import python as mp_python
from mediapipe.tasks.python import vision
from mediapipe import solutions
from mediapipe.framework.formats import landmark_pb2

from skellytracker.trackers.base_tracker.base_tracker import BaseTracker
from skellytracker.trackers.base_tracker.tracked_object import TrackedObject
from skellytracker.trackers.mediapipe_blendshape_tracker.mediapipe_blendshape_model_info import (
    MediapipeBlendshapeModelInfo,
)
from skellytracker.trackers.mediapipe_blendshape_tracker.mediapipe_blendshape_recorder import (
    MediapipeBlendshapeRecorder,
)


class MediapipeBlendshapeTracker(BaseTracker):
    def __init__(
        self,
        model_path: Union[Path, str, None] = None,
    ):
        super().__init__(
            tracked_object_names=MediapipeBlendshapeModelInfo.tracked_object_names,
            recorder=MediapipeBlendshapeRecorder(),
        )
        self.model_info = MediapipeBlendshapeModelInfo

        # If model_path not provided, try default model path, and if that doesn't work download model
        if model_path is None:
            model_path = self.get_or_download_mediapipe_blendshape_model()

        base_options = mp_python.BaseOptions(
            model_asset_path=str(model_path)
        )  # TODO: handle this path properly
        options = vision.FaceLandmarkerOptions(
            base_options=base_options,
            output_face_blendshapes=True,
            output_facial_transformation_matrixes=True,
            num_faces=1,
        )
        self.detector = vision.FaceLandmarker.create_from_options(options)

    def process_image(self, image: np.ndarray, **kwargs) -> Dict[str, TrackedObject]:
        rgb_image = cv2.cvtColor(
            image, cv2.COLOR_BGR2RGB
        )  # TODO: may need to convert this into an `mp.Image`, but can't find documentation about that

        mediapipe_image = mp.Image(image_format=mp.ImageFormat.SRGB, data=rgb_image)

        results = self.detector.detect(mediapipe_image)

        if len(results.face_blendshapes) == 0:
            self.annotated_image = image
            return {}

        self.tracked_objects["face"].extra["blendshapes"] = [
            blendshape.score for blendshape in results.face_blendshapes[0]
        ]  # TODO: assumes we're only interested in 1 face, but docs say this works for multiple faces??

        self.annotated_image = self.annotate_image(
            image=image,
            tracked_objects=self.tracked_objects,
            face_landmarks=results.face_landmarks[0],
        )

        return self.tracked_objects

    def annotate_image(
        self,
        image: np.ndarray,
        tracked_objects: Dict[str, TrackedObject],
        face_landmarks: list[landmark_pb2.NormalizedLandmark],
        **kwargs,
    ) -> np.ndarray:
        annotated_image = image.copy()

        face_landmarks_proto = landmark_pb2.NormalizedLandmarkList()
        face_landmarks_proto.landmark.extend(
            [
                landmark_pb2.NormalizedLandmark(
                    x=landmark.x, y=landmark.y, z=landmark.z
                )
                for landmark in face_landmarks
            ]
        )

        solutions.drawing_utils.draw_landmarks(
            image=annotated_image,
            landmark_list=face_landmarks_proto,
            connections=mp.solutions.face_mesh.FACEMESH_TESSELATION,
            landmark_drawing_spec=None,
            connection_drawing_spec=mp.solutions.drawing_styles.get_default_face_mesh_tesselation_style(),
        )
        solutions.drawing_utils.draw_landmarks(
            image=annotated_image,
            landmark_list=face_landmarks_proto,
            connections=mp.solutions.face_mesh.FACEMESH_CONTOURS,
            landmark_drawing_spec=None,
            connection_drawing_spec=mp.solutions.drawing_styles.get_default_face_mesh_contours_style(),
        )
        solutions.drawing_utils.draw_landmarks(
            image=annotated_image,
            landmark_list=face_landmarks_proto,
            connections=mp.solutions.face_mesh.FACEMESH_IRISES,
            landmark_drawing_spec=None,
            connection_drawing_spec=mp.solutions.drawing_styles.get_default_face_mesh_iris_connections_style(),
        )

        return annotated_image

    def get_or_download_mediapipe_blendshape_model(self) -> Path:
        model_path = Path(
            "face_landmarker_v2_with_blendshapes.task"
        )  # TODO: make this path absolute to skellytracker install
        if not model_path.exists():
            r = requests.get(
                "https://storage.googleapis.com/mediapipe-models/face_landmarker/face_landmarker/float16/1/face_landmarker.task",
                stream=True,
                timeout=(5, 60),
            )
            r.raise_for_status()
            model_path.write_bytes(r.content)
        return model_path
    

if __name__ == "__main__":
    MediapipeBlendshapeTracker().demo()



import cv2
import pytest
import numpy as np


from skellytracker.trackers.charuco_tracker.charuco_tracker import CharucoTracker


@pytest.mark.usefixtures("charuco_test_image")
def test_process_image(charuco_test_image):
    charuco_squares_x_in = 7
    charuco_squares_y_in = 5
    number_of_charuco_markers = (charuco_squares_x_in - 1) * (charuco_squares_y_in - 1)
    charuco_ids = [str(index) for index in range(number_of_charuco_markers)]

    tracker = CharucoTracker(
        tracked_object_names=charuco_ids,
        squares_x=charuco_squares_x_in,
        squares_y=charuco_squares_y_in,
        dictionary=cv2.aruco.getPredefinedDictionary(cv2.aruco.DICT_4X4_250),
    )
    tracked_objects = tracker.process_image(charuco_test_image)

    expected_results = {
        "0": (307.99796, 110.00571),
        "1": (336.93832, 120.001335),
        "2": (366.70923, 130.58067),
        "3": (396.36816, 141.78345),
        "4": (425.60236, 153.44989),
        "5": (455.322, 165.67236),
        "6": (294.39023, 135.20029),
        "7": (323.91107, 145.17943),
        "8": (353.4799, 155.70189),
        "9": (383.16925, 167.31921),
        "10": (412.6318, 179.24583),
        "11": (442.47086, 191.43738),
        "12": (280.9244, 160.91164),
        "13": (310.19556, 171.59216),
        "14": (339.86594, 182.31856),
        "15": (369.6749, 193.77588),
        "16": (399.5786, 205.84789),
        "17": (429.53903, 218.09035),
        "18": (267.12946, 187.75053),
        "19": (296.7608, 198.35608),
        "20": (326.31488, 209.67616),
        "21": (356.24832, 220.93948),
        "22": (386.3587, 233.16882),
        "23": (416.3993, 245.57997),
    }

    assert len(tracked_objects) == len(charuco_ids)
    for id in charuco_ids:
        tracked_corner = tracked_objects[id]
        assert tracked_corner.pixel_x is not None
        assert tracked_corner.pixel_y is not None
        assert np.allclose(
            (tracked_corner.pixel_x, tracked_corner.pixel_y), expected_results[id]
        )


@pytest.mark.usefixtures("test_image")
def test_image_without_charuco(test_image):
    charuco_squares_x_in = 7
    charuco_squares_y_in = 5
    number_of_charuco_markers = (charuco_squares_x_in - 1) * (charuco_squares_y_in - 1)
    charuco_ids = [str(index) for index in range(number_of_charuco_markers)]

    tracker = CharucoTracker(
        tracked_object_names=charuco_ids,
        squares_x=charuco_squares_x_in,
        squares_y=charuco_squares_y_in,
        dictionary=cv2.aruco.getPredefinedDictionary(cv2.aruco.DICT_4X4_250),
    )
    tracked_objects = tracker.process_image(test_image)

    assert len(tracked_objects) == len(charuco_ids)
    for id in charuco_ids:
        tracked_corner = tracked_objects[id]
        assert tracked_corner.pixel_x is None
        assert tracked_corner.pixel_y is None


@pytest.mark.usefixtures("charuco_test_image")
def test_annotate_image(charuco_test_image):
    charuco_squares_x_in = 7
    charuco_squares_y_in = 5
    number_of_charuco_markers = (charuco_squares_x_in - 1) * (charuco_squares_y_in - 1)
    charuco_ids = [str(index) for index in range(number_of_charuco_markers)]

    tracker = CharucoTracker(
        tracked_object_names=charuco_ids,
        squares_x=charuco_squares_x_in,
        squares_y=charuco_squares_y_in,
        dictionary=cv2.aruco.getPredefinedDictionary(cv2.aruco.DICT_4X4_250),
    )
    tracker.process_image(charuco_test_image)

    assert tracker.annotated_image is not None
    assert not np.all(tracker.annotated_image == charuco_test_image)


@pytest.mark.usefixtures("charuco_test_image")
def test_record(charuco_test_image):
    charuco_squares_x_in = 7
    charuco_squares_y_in = 5
    number_of_charuco_markers = (charuco_squares_x_in - 1) * (charuco_squares_y_in - 1)
    charuco_ids = [str(index) for index in range(number_of_charuco_markers)]

    tracker = CharucoTracker(
        tracked_object_names=charuco_ids,
        squares_x=charuco_squares_x_in,
        squares_y=charuco_squares_y_in,
        dictionary=cv2.aruco.getPredefinedDictionary(cv2.aruco.DICT_4X4_250),
    )
    tracked_objects = tracker.process_image(charuco_test_image)
    tracker.recorder.record(tracked_objects=tracked_objects)
    assert len(tracker.recorder.recorded_objects) == 1

    processed_results = tracker.recorder.process_tracked_objects()
    assert processed_results is not None
    assert processed_results.shape == (1, len(charuco_ids), 2)

    # expected_results = np.array(
    #     [
    #         [
    #             [392.56927490234375, 140.40118408203125, np.nan],
    #             [414.94940185546875, 132.90655517578125, np.nan],
    #             [386.4353332519531, 125.51483154296875, np.nan],
    #             [446.2061767578125, 157.98883056640625, np.nan],
    #             [373.6619873046875, 138.93646240234375, np.nan],
    #             [453.78662109375, 265.081787109375, np.nan],
    #             [317.9375305175781, 231.9653778076172, np.nan],
    #             [465.893310546875, 396.12274169921875, np.nan],
    #             [220.12176513671875, 325.96636962890625, np.nan],
    #             [465.23358154296875, 499.0487365722656, np.nan],
    #             [142.17066955566406, 407.7397155761719, np.nan],
    #             [352.325439453125, 468.44671630859375, np.nan],
    #             [268.8867492675781, 448.7227783203125, np.nan],
    #             [310.899658203125, 630.5478515625, np.nan],
    #             [227.7810821533203, 617.9011840820312, np.nan],
    #             [269.08587646484375, 733.4285888671875, np.nan],
    #             [213.1557159423828, 741.54541015625, np.nan],
    #         ]
    #     ]
    # )
    # assert np.allclose(processed_results[:, :, :2], expected_results[:, :, :2], atol=1e-2)
    # assert np.isnan(processed_results[:, :, 2]).all()



from skellytracker.trackers.base_tracker.base_tracking_params import BaseTrackingParams
from skellytracker.trackers.base_tracker.model_info import ModelInfo

from typing import Optional


class OpenPoseModelInfo(ModelInfo):
    name = "openpose"
    tracker_name = "OpenPoseTracker"
    body_landmark_names = [
        "nose",
        "neck",
        "right_shoulder",
        "right_elbow",
        "right_wrist",
        "left_shoulder",
        "left_elbow",
        "left_wrist",
        "hip_center",
        "right_hip",
        "right_knee",
        "right_ankle",
        "left_hip",
        "left_knee",
        "left_ankle",
        "right_eye",
        "left_eye", 
        "right_ear", 
        "left_ear", 
        "left_big_toe",
        "left_small_toe", 
        "left_heel", 
        "right_big_toe", 
        "right_small_toe", 
        "right_heel"
    ]
    landmark_names = body_landmark_names
    num_tracked_points_body = len(body_landmark_names)
    num_tracked_points_face = 70
    num_tracked_points_left_hand = 21
    num_tracked_points_right_hand = 21

    num_tracked_points = (
        num_tracked_points_body
        + num_tracked_points_left_hand
        + num_tracked_points_right_hand
        + num_tracked_points_face
    )
    tracked_object_names = ["pose_landmarks"]
    virtual_markers_definitions = {
        "head_center": {
            "marker_names": ["left_ear", "right_ear"],
            "marker_weights": [0.5, 0.5],
        },
        "trunk_center": {
            "marker_names": [
                "left_shoulder",
                "right_shoulder",
                "left_hip",
                "right_hip",
            ],
            "marker_weights": [0.25, 0.25, 0.25, 0.25],
        },
    }
    segment_connections = {
        "head": {"proximal": "left_ear", "distal": "right_ear"},
        "neck": {"proximal": "head_center", "distal": "neck"},
        "spine": {"proximal": "neck", "distal": "hip_center"},
        "right_shoulder": {"proximal": "neck", "distal": "right_shoulder"},
        "left_shoulder": {"proximal": "neck", "distal": "left_shoulder"},
        "right_upper_arm": {"proximal": "right_shoulder", "distal": "right_elbow"},
        "left_upper_arm": {"proximal": "left_shoulder", "distal": "left_elbow"},
        "right_forearm": {"proximal": "right_elbow", "distal": "right_wrist"},
        "left_forearm": {"proximal": "left_elbow", "distal": "left_wrist"},
        "right_pelvis": {"proximal": "hip_center", "distal": "right_hip"},
        "left_pelvis": {"proximal": "hip_center", "distal": "left_hip"},
        "right_thigh": {"proximal": "right_hip", "distal": "right_knee"},
        "left_thigh": {"proximal": "left_hip", "distal": "left_knee"},
        "right_shank": {"proximal": "right_knee", "distal": "right_ankle"},
        "left_shank": {"proximal": "left_knee", "distal": "left_ankle"},
        "right_foot": {"proximal": "right_ankle", "distal": "right_big_toe"},
        "left_foot": {"proximal": "left_ankle", "distal": "left_big_toe"},
        "right_heel": {"proximal": "right_ankle", "distal": "right_heel"},
        "left_heel": {"proximal": "left_ankle", "distal": "left_heel"},
        "right_foot_bottom": {"proximal": "right_heel", "distal": "right_big_toe"},
        "left_foot_bottom": {"proximal": "left_heel", "distal": "left_big_toe"},
    }
    center_of_mass_definitions = { #NOTE: using forearm/hand definition from Winter tables, as we don't have hand definitions here  
        "head": {
            "segment_com_length": .5,
            "segment_com_percentage": .081,
        },
        "spine": {
            "segment_com_length": 0.5,
            "segment_com_percentage": 0.497,
        },
        "right_upper_arm": {
            "segment_com_length": 0.436,
            "segment_com_percentage": 0.028,
        },
        "left_upper_arm": {
            "segment_com_length": 0.436,
            "segment_com_percentage": 0.028,
        },
        "right_forearm": {
            "segment_com_length": 0.682,
            "segment_com_percentage": 0.022,
        },
        "left_forearm": {
            "segment_com_length": 0.682,
            "segment_com_percentage": 0.022,
        },
        "right_thigh": {
            "segment_com_length": 0.433,
            "segment_com_percentage": 0.1,
        },
        "left_thigh": {
            "segment_com_length": 0.433,
            "segment_com_percentage": 0.1,
        },
        "right_shank": {
            "segment_com_length": 0.433,
            "segment_com_percentage": 0.0465,
        },
        "left_shank": {
            "segment_com_length": 0.433,
            "segment_com_percentage": 0.0465,
        },
        "right_foot": {
            "segment_com_length": 0.5,
            "segment_com_percentage": 0.0145,
        },
        "left_foot": {
            "segment_com_length": 0.5,
            "segment_com_percentage": 0.0145,
        },
    }        
    joint_hierarchy = {
        "hip_center": ["left_hip", "right_hip", "trunk_center"],
        "trunk_center": ["neck"],
        "neck": ["left_shoulder", "right_shoulder", "head_center"],
        "head_center": ["nose", "left_ear", "right_ear", "left_eye", "right_eye"],
        "left_shoulder": ["left_elbow"],
        "left_elbow": ["left_wrist"],
        "right_shoulder": ["right_elbow"],
        "right_elbow": ["right_wrist"],
        "left_hip": ["left_knee"],
        "left_knee": ["left_ankle"],
        "left_ankle": ["left_big_toe", "left_small_toe", "left_heel"],
        "right_hip": ["right_knee"],
        "right_knee": ["right_ankle"],
        "right_ankle": ["right_big_toe", "right_small_toe", "right_heel"],
    }


class OpenPoseTrackingParams(BaseTrackingParams):
    openpose_root_folder_path: str
    output_json_path: Optional[str] = None
    net_resolution: str = "-1x320"
    number_people_max: int = 1
    track_hands: bool = True
    track_face: bool = True
    write_video: bool = True
    output_resolution: str = "-1x-1"



import logging

import cv2
import numpy as np
import requests

from skellytracker.system.default_paths import FIGSHARE_TEST_IMAGE_URL


logger = logging.getLogger(__name__)


def download_test_image(test_image_url: str = FIGSHARE_TEST_IMAGE_URL) -> np.ndarray:
    try:
        logger.info(f"Downloading test image from {test_image_url}...")

        r = requests.get(test_image_url, stream=True, timeout=(5, 60))
        r.raise_for_status()  # Check if request was successful

        image_array = np.frombuffer(r.content, np.uint8)
        image = cv2.imdecode(image_array, cv2.IMREAD_COLOR)

        logger.info("Test image downloaded successfully.")
        return image

    except requests.exceptions.RequestException as e:
        logger.error(f"Request failed: {e}")
        raise e


if __name__ == "__main__":
    test_data_path = download_test_image()



import numpy as np
import pytest
from skellytracker.system.default_paths import FIGSHARE_CHARUCO_TEST_IMAGE_URL
from skellytracker.utilities.download_test_image import download_test_image


class SessionInfo:
    test_image: np.ndarray
    charuco_test_image: np.ndarray


def pytest_sessionstart(session):
    SessionInfo.test_image = download_test_image()
    SessionInfo.charuco_test_image = download_test_image(test_image_url=FIGSHARE_CHARUCO_TEST_IMAGE_URL)


@pytest.fixture()
def test_image():
    return SessionInfo.test_image

@pytest.fixture
def charuco_test_image():
    return SessionInfo.charuco_test_image
